@@ -1,13 +1,24 @@
 # frozen_string_literal: true
 
 require_relative 'helper'
-require 'cgi/escape'
+require 'cgi'
 require 'forwardable'
 require 'securerandom'
 
+separate_testing do
+  require_relative '../lib/rack/request'
+  require_relative '../lib/rack/mock_request'
+  require_relative '../lib/rack/lint'
+end
+
 class RackRequestTest < Minitest::Spec
   it "copies the env when duping" do
     req = make_request(Rack::MockRequest.env_for("http://example.com:8080/"))
+
+    if req.delegate?
+      skip "delegate requests don't dup environments"
+    end
+
     refute_same req.env, req.dup.env
   end
 
@@ -36,6 +47,20 @@ class RackRequestTest < Minitest::Spec
     assert_equal "example.com:443", req.authority
   end
 
+  it 'can calculate the server authority' do
+    req = make_request('SERVER_NAME' => 'example.com')
+    assert_equal "example.com", req.server_authority
+    req = make_request('SERVER_NAME' => 'example.com', 'SERVER_PORT' => 8080)
+    assert_equal "example.com:8080", req.server_authority
+  end
+
+  it 'can calculate the port without an authority' do
+    req = make_request('SERVER_PORT' => 8080)
+    assert_equal 8080, req.port
+    req = make_request('HTTPS' => 'on')
+    assert_equal 443, req.port
+  end
+
   it 'yields to the block if no value has been set' do
     req = make_request(Rack::MockRequest.env_for("http://example.com:8080/"))
     yielded = false
@@ -122,15 +147,67 @@ class RackRequestTest < Minitest::Spec
     req.hostname.must_equal "123foo.example.com"
 
     req = make_request \
-      Rack::MockRequest.env_for("/", "HTTP_HOST" => "some_service:3001")
-    req.host.must_equal "some_service"
-    req.hostname.must_equal "some_service"
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "♡.com")
+    req.host.must_equal "♡.com"
+    req.hostname.must_equal "♡.com"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "♡.com:80")
+    req.host.must_equal "♡.com"
+    req.hostname.must_equal "♡.com"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "nic.谷歌")
+    req.host.must_equal "nic.谷歌"
+    req.hostname.must_equal "nic.谷歌"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "nic.谷歌:80")
+    req.host.must_equal "nic.谷歌"
+    req.hostname.must_equal "nic.谷歌"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "technically_invalid.example.com")
+    req.host.must_equal "technically_invalid.example.com"
+    req.hostname.must_equal "technically_invalid.example.com"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "technically_invalid.example.com:80")
+    req.host.must_equal "technically_invalid.example.com"
+    req.hostname.must_equal "technically_invalid.example.com"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "trailing_newline.com\n")
+    req.host.must_be_nil
+    req.hostname.must_be_nil
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "really\nbad\ninput")
+    req.host.must_be_nil
+    req.hostname.must_be_nil
 
     req = make_request \
       Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org", "SERVER_PORT" => "9292")
     req.host.must_equal "example.org"
     req.hostname.must_equal "example.org"
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_FORWARDED" => "host=example.org:9292")
+    req.host.must_equal "example.org"
+
+    # Test obfuscated identifier: https://tools.ietf.org/html/rfc7239#section-6.3
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_FORWARDED" => "host=ObFuScaTeD")
+    req.host.must_equal "ObFuScaTeD"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_FORWARDED" => "host=example.com; host=example.org:9292")
+    req.host.must_equal "example.org"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org:9292", "HTTP_FORWARDED" => "host=example.com")
+    req.host.must_equal "example.com"
+
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org:9292")
     req.host.must_equal "example.org"
@@ -146,6 +223,71 @@ class RackRequestTest < Minitest::Spec
     req.host.must_equal "[2001:db8:cafe::17]"
     req.hostname.must_equal "2001:db8:cafe::17"
 
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[::]:47011")
+    req.host.must_equal "[::]"
+    req.hostname.must_equal "::"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[1111:2222:3333:4444:5555:6666:123.123.123.123]")
+    req.host.must_equal "[1111:2222:3333:4444:5555:6666:123.123.123.123]"
+    req.hostname.must_equal "1111:2222:3333:4444:5555:6666:123.123.123.123"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[1111:2222:3333:4444:5555:6666:123.123.123.123]:47011")
+    req.host.must_equal "[1111:2222:3333:4444:5555:6666:123.123.123.123]"
+    req.hostname.must_equal "1111:2222:3333:4444:5555:6666:123.123.123.123"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "0.0.0.0")
+    req.host.must_equal "0.0.0.0"
+    req.hostname.must_equal "0.0.0.0"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "0.0.0.0:47011")
+    req.host.must_equal "0.0.0.0"
+    req.hostname.must_equal "0.0.0.0"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "255.255.255.255")
+    req.host.must_equal "255.255.255.255"
+    req.hostname.must_equal "255.255.255.255"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "255.255.255.255:47011")
+    req.host.must_equal "255.255.255.255"
+    req.hostname.must_equal "255.255.255.255"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "really\nbad\ninput")
+    req.host.must_be_nil
+    req.hostname.must_be_nil
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[0]")
+    req.host.must_be_nil
+    req.hostname.must_be_nil
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[:::]")
+    req.host.must_be_nil
+    req.hostname.must_be_nil
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "[1111:2222:3333:4444:5555:6666:7777:88888]")
+    req.host.must_be_nil
+    req.hostname.must_be_nil
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "0.0..0.0")
+    req.host.must_equal '0.0..0.0'
+    req.hostname.must_equal '0.0..0.0'
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "255.255.255.0255")
+    req.host.must_equal "255.255.255.0255"
+    req.hostname.must_equal "255.255.255.0255"
+
     env = Rack::MockRequest.env_for("/")
     env.delete("SERVER_NAME")
     req = make_request(env)
@@ -161,10 +303,6 @@ class RackRequestTest < Minitest::Spec
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "www2.example.org:81")
     req.port.must_equal 81
 
-    req = make_request \
-      Rack::MockRequest.env_for("/", "HTTP_HOST" => "some_service:3001")
-    req.port.must_equal 3001
-
     req = make_request \
       Rack::MockRequest.env_for("/", "SERVER_NAME" => "example.org", "SERVER_PORT" => "9292")
     req.port.must_equal 9292
@@ -212,6 +350,187 @@ class RackRequestTest < Minitest::Spec
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost", "HTTP_X_FORWARDED_PROTO" => "https,https", "SERVER_PORT" => "80")
     req.port.must_equal 443
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost", "HTTP_FORWARDED" => "proto=https", "HTTP_X_FORWARDED_PROTO" => "http", "SERVER_PORT" => "9393")
+    req.port.must_equal 443
+  end
+
+  it "have forwarded_* methods respect forwarded_priority" do
+    begin
+      default_priority = Rack::Request.forwarded_priority
+      default_proto_priority = Rack::Request.x_forwarded_proto_priority
+
+      def self.req(headers)
+        req = make_request Rack::MockRequest.env_for("/", headers)
+        req.singleton_class.send(:public, :forwarded_scheme)
+        req
+      end
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4",
+        "HTTP_X_FORWARDED_FOR" => "2.3.4.5").
+        forwarded_for.must_equal ['1.2.3.4']
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4:1234",
+        "HTTP_X_FORWARDED_PORT" => "2345").
+        forwarded_port.must_equal [1234]
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4",
+        "HTTP_X_FORWARDED_PORT" => "2345").
+        forwarded_port.must_equal []
+
+      req("HTTP_FORWARDED"=>"host=1.2.3.4, host=3.4.5.6",
+        "HTTP_X_FORWARDED_HOST" => "2.3.4.5,4.5.6.7").
+        forwarded_authority.must_equal '3.4.5.6'
+
+      req("HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "ws"
+
+      req("HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "http"
+
+      Rack::Request.forwarded_priority = [nil, :x_forwarded, :forwarded]
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4",
+        "HTTP_X_FORWARDED_FOR" => "2.3.4.5").
+        forwarded_for.must_equal ['2.3.4.5']
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4",
+        "HTTP_X_FORWARDED_PORT" => "2345").
+        forwarded_port.must_equal [2345]
+
+      req("HTTP_FORWARDED"=>"host=1.2.3.4, host=3.4.5.6",
+        "HTTP_X_FORWARDED_HOST" => "2.3.4.5,4.5.6.7").
+        forwarded_authority.must_equal '4.5.6.7'
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "ws"
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "http"
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_equal "https"
+
+      Rack::Request.x_forwarded_proto_priority = [nil, :scheme, :proto]
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "http"
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws").
+        forwarded_scheme.must_equal "ws"
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_equal "https"
+
+      Rack::Request.forwarded_priority = [:x_forwarded]
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "http"
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws").
+        forwarded_scheme.must_equal "ws"
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_be_nil
+
+      Rack::Request.x_forwarded_proto_priority = [:scheme]
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "http"
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_be_nil
+
+      Rack::Request.x_forwarded_proto_priority = [:proto]
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal "ws"
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_be_nil
+
+      Rack::Request.x_forwarded_proto_priority = []
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_be_nil
+
+      Rack::Request.x_forwarded_proto_priority = default_proto_priority
+      Rack::Request.forwarded_priority = [:forwarded]
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_equal 'https'
+
+      req("HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_X_FORWARDED_PROTO" => "ws").
+        forwarded_scheme.must_be_nil
+
+      Rack::Request.forwarded_priority = []
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4",
+        "HTTP_X_FORWARDED_FOR" => "2.3.4.5").
+        forwarded_for.must_be_nil
+
+      req("HTTP_FORWARDED"=>"for=1.2.3.4",
+        "HTTP_X_FORWARDED_PORT" => "2345").
+        forwarded_port.must_be_nil
+
+      req("HTTP_FORWARDED"=>"host=1.2.3.4, host=3.4.5.6",
+        "HTTP_X_FORWARDED_HOST" => "2.3.4.5,4.5.6.7").
+        forwarded_authority.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_PROTO" => "ws",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https",
+        "HTTP_X_FORWARDED_SCHEME" => "http").
+        forwarded_scheme.must_be_nil
+
+      req("HTTP_FORWARDED"=>"proto=https").
+        forwarded_scheme.must_be_nil
+
+    ensure
+      Rack::Request.forwarded_priority = default_priority
+      Rack::Request.x_forwarded_proto_priority = default_proto_priority
+    end
   end
 
   it "figure out the correct host with port" do
@@ -246,14 +565,18 @@ class RackRequestTest < Minitest::Spec
     req = make_request \
       Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org", "SERVER_PORT" => "9393")
     req.host_with_port.must_equal "example.org"
+
+    req = make_request \
+      Rack::MockRequest.env_for("/", "HTTP_HOST" => "localhost:81", "HTTP_X_FORWARDED_HOST" => "example.org", "HTTP_FORWARDED" => "host=example.com:9292", "SERVER_PORT" => "9393")
+    req.host_with_port.must_equal "example.com:9292"
   end
 
   it "parse the query string" do
-    req = make_request(Rack::MockRequest.env_for("/?foo=bar&quux=bla"))
-    req.query_string.must_equal "foo=bar&quux=bla"
-    req.GET.must_equal "foo" => "bar", "quux" => "bla"
-    req.POST.must_be :empty?
-    req.params.must_equal "foo" => "bar", "quux" => "bla"
+    request = make_request(Rack::MockRequest.env_for("/?foo=bar&quux=bla&nothing&empty="))
+    request.query_string.must_equal "foo=bar&quux=bla&nothing&empty="
+    request.GET.must_equal "foo" => "bar", "quux" => "bla", "nothing" => nil, "empty" => ""
+    request.POST.must_be :empty?
+    request.params.must_equal "foo" => "bar", "quux" => "bla", "nothing" => nil, "empty" => ""
   end
 
   it "not truncate query strings containing semi-colons #543 only in POST" do
@@ -274,7 +597,7 @@ class RackRequestTest < Minitest::Spec
         @params = Hash.new{|h, k| h[k.to_s] if k.is_a?(Symbol)}
       end
     end
-    parser = Rack::QueryParser.new(c, 65536, 100)
+    parser = Rack::QueryParser.new(c, 100)
     c = Class.new(Rack::Request) do
       define_method(:query_parser) do
         parser
@@ -287,49 +610,23 @@ class RackRequestTest < Minitest::Spec
     req.params[:quux].must_equal "bla"
   end
 
-  it "use semi-colons as separators for query strings in GET" do
+  it "does not use semi-colons as separators for query strings in GET" do
     req = make_request(Rack::MockRequest.env_for("/?foo=bar&quux=b;la;wun=duh"))
     req.query_string.must_equal "foo=bar&quux=b;la;wun=duh"
-    req.GET.must_equal "foo" => "bar", "quux" => "b", "la" => nil, "wun" => "duh"
+    req.GET.must_equal "foo" => "bar", "quux" => "b;la;wun=duh"
     req.POST.must_be :empty?
-    req.params.must_equal "foo" => "bar", "quux" => "b", "la" => nil, "wun" => "duh"
-  end
-
-  it "limit the keys from the GET query string" do
-    env = Rack::MockRequest.env_for("/?foo=bar")
-
-    old, Rack::Utils.key_space_limit = Rack::Utils.key_space_limit, 1
-    begin
-      req = make_request(env)
-      lambda { req.GET }.must_raise Rack::QueryParser::ParamsTooDeepError
-    ensure
-      Rack::Utils.key_space_limit = old
-    end
-  end
-
-  it "limit the key size per nested params hash" do
-    nested_query = Rack::MockRequest.env_for("/?foo%5Bbar%5D%5Bbaz%5D%5Bqux%5D=1")
-    plain_query  = Rack::MockRequest.env_for("/?foo_bar__baz__qux_=1")
-
-    old, Rack::Utils.key_space_limit = Rack::Utils.key_space_limit, 3
-    begin
-      exp = { "foo" => { "bar" => { "baz" => { "qux" => "1" } } } }
-      make_request(nested_query).GET.must_equal exp
-      lambda { make_request(plain_query).GET  }.must_raise Rack::QueryParser::ParamsTooDeepError
-    ensure
-      Rack::Utils.key_space_limit = old
-    end
+    req.params.must_equal "foo" => "bar", "quux" => "b;la;wun=duh"
   end
 
   it "limit the allowed parameter depth when parsing parameters" do
-    env = Rack::MockRequest.env_for("/?a#{'[a]' * 110}=b")
+    env = Rack::MockRequest.env_for("/?a#{'[a]' * 40}=b")
     req = make_request(env)
     lambda { req.GET }.must_raise Rack::QueryParser::ParamsTooDeepError
 
-    env = Rack::MockRequest.env_for("/?a#{'[a]' * 90}=b")
+    env = Rack::MockRequest.env_for("/?a#{'[a]' * 30}=b")
     req = make_request(env)
     params = req.GET
-    90.times { params = params['a'] }
+    30.times { params = params['a'] }
     params['a'].must_equal 'b'
 
     old, Rack::Utils.param_depth_limit = Rack::Utils.param_depth_limit, 3
@@ -367,7 +664,7 @@ class RackRequestTest < Minitest::Spec
         @params = Hash.new{|h, k| h[k.to_s] if k.is_a?(Symbol)}
       end
     end
-    parser = Rack::QueryParser.new(c, 65536, 100)
+    parser = Rack::QueryParser.new(c, 100)
     c = Class.new(Rack::Request) do
       define_method(:query_parser) do
         parser
@@ -399,12 +696,7 @@ class RackRequestTest < Minitest::Spec
       message.must_equal "invalid %-encoding (a%)"
   end
 
-  it "raise if rack.input is missing" do
-    req = make_request({})
-    lambda { req.POST }.must_raise RuntimeError
-  end
-
-  it "parse POST data when method is POST and no Content-Type given" do
+  it "parse POST data when method is POST and no content-type given" do
     req = make_request \
       Rack::MockRequest.env_for("/?foo=quux",
         "REQUEST_METHOD" => 'POST',
@@ -417,20 +709,6 @@ class RackRequestTest < Minitest::Spec
     req.params.must_equal "foo" => "bar", "quux" => "bla"
   end
 
-  it "limit the keys from the POST form data" do
-    env = Rack::MockRequest.env_for("",
-            "REQUEST_METHOD" => 'POST',
-            :input => "foo=bar&quux=bla")
-
-    old, Rack::Utils.key_space_limit = Rack::Utils.key_space_limit, 1
-    begin
-      req = make_request(env)
-      lambda { req.POST }.must_raise Rack::QueryParser::ParamsTooDeepError
-    ensure
-      Rack::Utils.key_space_limit = old
-    end
-  end
-
   it "parse POST data with explicit content type regardless of method" do
     req = make_request \
       Rack::MockRequest.env_for("/",
@@ -453,7 +731,9 @@ class RackRequestTest < Minitest::Spec
     req.media_type.must_equal 'text/plain'
     req.media_type_params['charset'].must_equal 'utf-8'
     req.content_charset.must_equal 'utf-8'
-    req.POST.must_be :empty?
+    post = req.POST
+    post.must_be_empty
+    req.POST.must_be_same_as post
     req.params.must_equal "foo" => "quux"
     req.body.read.must_equal "foo=bar&quux=bla"
   end
@@ -465,17 +745,6 @@ class RackRequestTest < Minitest::Spec
         "CONTENT_TYPE" => 'application/x-www-form-urlencoded',
         :input => "foo=bar&quux=bla")
     req.POST.must_equal "foo" => "bar", "quux" => "bla"
-    req.body.read.must_equal "foo=bar&quux=bla"
-  end
-
-  it "rewind input after parsing POST data" do
-    input = StringIO.new("foo=bar&quux=bla")
-    req = make_request \
-      Rack::MockRequest.env_for("/",
-        "CONTENT_TYPE" => 'application/x-www-form-urlencoded;foo=bar',
-        :input => input)
-    req.params.must_equal "foo" => "bar", "quux" => "bla"
-    input.read.must_equal "foo=bar&quux=bla"
   end
 
   it "safely accepts POST requests with empty body" do
@@ -502,19 +771,21 @@ class RackRequestTest < Minitest::Spec
   it "get value by key from params with #[]" do
     req = make_request \
       Rack::MockRequest.env_for("?foo=quux")
-    req['foo'].must_equal 'quux'
-    req[:foo].must_equal 'quux'
+    assert_output(nil, /deprecated/) do
+      req['foo'].must_equal 'quux'
+      req[:foo].must_equal 'quux'
+    end
 
     next if self.class == TestProxyRequest
     verbose = $VERBOSE
     warn_arg = nil
-    req.define_singleton_method(:warn) do |arg|
-      warn_arg = arg
+    req.define_singleton_method(:warn) do |*args|
+      warn_arg = args
     end
     begin
       $VERBOSE = true
       req['foo'].must_equal 'quux'
-      warn_arg.must_equal "Request#[] is deprecated and will be removed in a future version of Rack. Please use request.params[] instead"
+      warn_arg.must_equal ["Request#[] is deprecated and will be removed in a future version of Rack. Please use request.params[] instead", { uplevel: 1 }]
     ensure
       $VERBOSE = verbose
     end
@@ -523,33 +794,43 @@ class RackRequestTest < Minitest::Spec
   it "set value to key on params with #[]=" do
     req = make_request \
       Rack::MockRequest.env_for("?foo=duh")
-    req['foo'].must_equal 'duh'
-    req[:foo].must_equal 'duh'
+    assert_output(nil, /deprecated/) do
+      req['foo'].must_equal 'duh'
+      req[:foo].must_equal 'duh'
+    end
     req.params.must_equal 'foo' => 'duh'
 
     if req.delegate?
       skip "delegate requests don't cache params, so mutations have no impact"
     end
 
-    req['foo'] = 'bar'
+    assert_output(nil, /deprecated/) do
+      req['foo'] = 'bar'
+    end
     req.params.must_equal 'foo' => 'bar'
-    req['foo'].must_equal 'bar'
-    req[:foo].must_equal 'bar'
+    assert_output(nil, /deprecated/) do
+      req['foo'].must_equal 'bar'
+      req[:foo].must_equal 'bar'
+    end
 
-    req[:foo] = 'jaz'
+    assert_output(nil, /deprecated/) do
+      req[:foo] = 'jaz'
+    end
     req.params.must_equal 'foo' => 'jaz'
-    req['foo'].must_equal 'jaz'
-    req[:foo].must_equal 'jaz'
+    assert_output(nil, /deprecated/) do
+      req['foo'].must_equal 'jaz'
+      req[:foo].must_equal 'jaz'
+    end
 
     verbose = $VERBOSE
     warn_arg = nil
-    req.define_singleton_method(:warn) do |arg|
-      warn_arg = arg
+    req.define_singleton_method(:warn) do |*args|
+      warn_arg = args
     end
     begin
       $VERBOSE = true
       req['foo'] = 'quux'
-      warn_arg.must_equal "Request#[]= is deprecated and will be removed in a future version of Rack. Please use request.params[]= instead"
+      warn_arg.must_equal ["Request#[]= is deprecated and will be removed in a future version of Rack. Please use request.params[]= instead", { uplevel: 1 }]
       req.params['foo'].must_equal 'quux'
     ensure
       $VERBOSE = verbose
@@ -634,6 +915,25 @@ class RackRequestTest < Minitest::Spec
     request.scheme.must_equal "http"
     request.wont_be :ssl?
 
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_X_FORWARDED_SCHEME' => 'ws'))
+    request.scheme.must_equal "ws"
+    request.wont_be :ssl?
+
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_X_FORWARDED_PROTO' => 'ws'))
+    request.scheme.must_equal "ws"
+
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_FORWARDED' => 'proto=https'))
+    request.scheme.must_equal "https"
+    request.must_be :ssl?
+
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_FORWARDED' => 'proto=https, proto=http'))
+    request.scheme.must_equal "http"
+    request.wont_be :ssl?
+
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_FORWARDED' => 'proto=http, proto=https'))
+    request.scheme.must_equal "https"
+    request.must_be :ssl?
+
     request = make_request(Rack::MockRequest.env_for("/", 'HTTPS' => 'on'))
     request.scheme.must_equal "https"
     request.must_be :ssl?
@@ -662,12 +962,20 @@ class RackRequestTest < Minitest::Spec
     request.scheme.must_equal "https"
     request.must_be :ssl?
 
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_X_FORWARDED_SCHEME' => 'wss'))
+    request.scheme.must_equal "wss"
+    request.must_be :ssl?
+
     request = make_request(Rack::MockRequest.env_for("/", 'HTTP_X_FORWARDED_PROTO' => 'https'))
     request.scheme.must_equal "https"
     request.must_be :ssl?
 
     request = make_request(Rack::MockRequest.env_for("/", 'HTTP_X_FORWARDED_PROTO' => 'https, http, http'))
-    request.scheme.must_equal "https"
+    request.scheme.must_equal "http"
+    request.wont_be :ssl?
+
+    request = make_request(Rack::MockRequest.env_for("/", 'HTTP_X_FORWARDED_PROTO' => 'wss'))
+    request.scheme.must_equal "wss"
     request.must_be :ssl?
   end
 
@@ -905,6 +1213,22 @@ class RackRequestTest < Minitest::Spec
       req.media_type_params['weird'].must_equal 'lol"'
   end
 
+  it "returns the same error for invalid post inputs" do
+    env = {
+      'REQUEST_METHOD' => 'POST',
+      'PATH_INFO' => '/foo',
+      'rack.input' => StringIO.new('invalid=bar&invalid[foo]=bar'),
+      'HTTP_CONTENT_TYPE' => "application/x-www-form-urlencoded",
+    }
+    
+    2.times do
+      # The actual exception type here is unimportant - just that it fails.
+      assert_raises(Rack::Utils::ParameterTypeError) do
+        Rack::Request.new(env).POST
+      end
+    end
+  end
+
   it "parse with junk before boundary" do
     # Adapted from RFC 1867.
     input = <<EOF
@@ -916,8 +1240,8 @@ content-disposition: form-data; name="reply"\r
 yes\r
 --AaB03x\r
 content-disposition: form-data; name="fileupload"; filename="dj.jpg"\r
-Content-Type: image/jpeg\r
-Content-Transfer-Encoding: base64\r
+content-type: image/jpeg\r
+content-transfer-encoding: base64\r
 \r
 /9j/4AAQSkZJRgABAQAAAQABAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcg\r
 --AaB03x--\r
@@ -955,8 +1279,8 @@ content-disposition: form-data; name="reply"
 yes
 --AaB03x
 content-disposition: form-data; name="fileupload"; filename="dj.jpg"
-Content-Type: image/jpeg
-Content-Transfer-Encoding: base64
+content-type: image/jpeg
+content-transfer-encoding: base64
 
 /9j/4AAQSkZJRgABAQAAAQABAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcg
 --AaB03x--
@@ -979,8 +1303,8 @@ content-disposition: form-data; name="reply"\r
 yes\r
 --AaB03x\r
 content-disposition: form-data; name="fileupload"; filename="dj.jpg"\r
-Content-Type: image/jpeg\r
-Content-Transfer-Encoding: base64\r
+content-type: image/jpeg\r
+content-transfer-encoding: base64\r
 \r
 /9j/4AAQSkZJRgABAQAAAQABAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcg\r
 --AaB03x--\r
@@ -1029,7 +1353,7 @@ EOF
 
   it "MultipartPartLimitError when request has too many multipart file parts if limit set" do
     begin
-      data = 10000.times.map { "--AaB03x\r\nContent-Type: text/plain\r\nContent-Disposition: attachment; name=#{SecureRandom.hex(10)}; filename=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
+      data = 10000.times.map { "--AaB03x\r\ncontent-type: text/plain\r\ncontent-disposition: attachment; name=#{SecureRandom.hex(10)}; filename=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
       data += "--AaB03x--\r"
 
       options = {
@@ -1061,7 +1385,7 @@ EOF
 
   it 'closes tempfiles it created in the case of too many created' do
     begin
-      data = 10000.times.map { "--AaB03x\r\nContent-Type: text/plain\r\nContent-Disposition: attachment; name=#{SecureRandom.hex(10)}; filename=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
+      data = 10000.times.map { "--AaB03x\r\ncontent-type: text/plain\r\ncontent-disposition: attachment; name=#{SecureRandom.hex(10)}; filename=#{SecureRandom.hex(10)}\r\n\r\ncontents\r\n" }.join("\r\n")
       data += "--AaB03x--\r"
 
       files = []
@@ -1111,14 +1435,14 @@ EOF
     input = <<EOF
 --AaB03x\r
 content-disposition: form-data; name="fileupload"; filename="foo.jpg"\r
-Content-Type: image/jpeg\r
-Content-Transfer-Encoding: base64\r
+content-type: image/jpeg\r
+content-transfer-encoding: base64\r
 \r
 /9j/4AAQSkZJRgABAQAAAQABAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcg\r
 --AaB03x\r
 content-disposition: form-data; name="fileupload"; filename="bar.jpg"\r
-Content-Type: image/jpeg\r
-Content-Transfer-Encoding: base64\r
+content-type: image/jpeg\r
+content-transfer-encoding: base64\r
 \r
 /9j/4AAQSkZJRgABAQAAAQABAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcg\r
 --AaB03x--\r
@@ -1188,9 +1512,9 @@ EOF
   it "correctly parse the part name from Content-Id header" do
     input = <<EOF
 --AaB03x\r
-Content-Type: text/xml; charset=utf-8\r
+content-type: text/xml; charset=utf-8\r
 Content-Id: <soap-start>\r
-Content-Transfer-Encoding: 7bit\r
+content-transfer-encoding: 7bit\r
 \r
 foo\r
 --AaB03x--\r
@@ -1228,19 +1552,23 @@ EOF
     rack_input.write(input)
     rack_input.rewind
 
-    req = make_request Rack::MockRequest.env_for("/",
-                      "rack.request.form_hash" => { 'foo' => 'bar' },
-                      "rack.request.form_input" => rack_input,
-                      :input => rack_input)
+    form_hash = {}
 
-    req.POST.must_equal req.env['rack.request.form_hash']
+    req = make_request Rack::MockRequest.env_for(
+      "/",
+      "rack.request.form_hash" => form_hash,
+      "rack.request.form_input" => rack_input,
+      :input => rack_input
+    )
+
+    req.POST.must_be_same_as form_hash
   end
 
   it "conform to the Rack spec" do
     app = lambda { |env|
       content = make_request(env).POST["file"].inspect
       size = content.bytesize
-      [200, { "Content-Type" => "text/html", "Content-Length" => size.to_s }, [content]]
+      [200, { "content-type" => "text/html", "content-length" => size.to_s }, [content]]
     }
 
     input = <<EOF.dup
@@ -1250,8 +1578,8 @@ content-disposition: form-data; name="reply"\r
 yes\r
 --AaB03x\r
 content-disposition: form-data; name="fileupload"; filename="dj.jpg"\r
-Content-Type: image/jpeg\r
-Content-Transfer-Encoding: base64\r
+content-type: image/jpeg\r
+content-transfer-encoding: base64\r
 \r
 /9j/4AAQSkZJRgABAQAAAQABAAD//gA+Q1JFQVRPUjogZ2QtanBlZyB2MS4wICh1c2luZyBJSkcg\r
 --AaB03x--\r
@@ -1317,7 +1645,7 @@ EOF
     res.body.must_equal 'fe80::202:b3ff:fe1e:8329'
 
     res = mock.get '/', 'REMOTE_ADDR' => '1.2.3.4,3.4.5.6'
-    res.body.must_equal '1.2.3.4'
+    res.body.must_equal '3.4.5.6'
 
     res = mock.get '/', 'REMOTE_ADDR' => '127.0.0.1'
     res.body.must_equal '127.0.0.1'
@@ -1329,6 +1657,21 @@ EOF
   it 'deals with proxies' do
     mock = Rack::MockRequest.new(Rack::Lint.new(ip_app))
 
+    res = mock.get '/',
+      'REMOTE_ADDR' => '1.2.3.4',
+      'HTTP_FORWARDED' => 'for=3.4.5.6'
+    res.body.must_equal '1.2.3.4'
+
+    res = mock.get '/',
+      'HTTP_X_FORWARDED_FOR' => '3.4.5.6',
+      'HTTP_FORWARDED' => 'for=5.6.7.8'
+    res.body.must_equal '5.6.7.8'
+
+    res = mock.get '/',
+      'HTTP_X_FORWARDED_FOR' => '3.4.5.6',
+      'HTTP_FORWARDED' => 'for=5.6.7.8, for=7.8.9.0'
+    res.body.must_equal '7.8.9.0'
+
     res = mock.get '/',
       'REMOTE_ADDR' => '1.2.3.4',
       'HTTP_X_FORWARDED_FOR' => '3.4.5.6'
@@ -1363,6 +1706,9 @@ EOF
     res = mock.get '/', 'HTTP_X_FORWARDED_FOR' => '[2001:db8:cafe::17]:47011'
     res.body.must_equal '2001:db8:cafe::17'
 
+    res = mock.get '/', 'HTTP_FORWARDED' => 'for="[2001:db8:cafe::17]:47011"'
+    res.body.must_equal '2001:db8:cafe::17'
+
     res = mock.get '/', 'HTTP_X_FORWARDED_FOR' => '1.2.3.4, [2001:db8:cafe::17]:47011'
     res.body.must_equal '2001:db8:cafe::17'
 
@@ -1458,14 +1804,21 @@ EOF
   it "regards local addresses as proxies" do
     req = make_request(Rack::MockRequest.env_for("/"))
     req.trusted_proxy?('127.0.0.1').must_equal true
+    req.trusted_proxy?('127.000.000.001').must_equal true
+    req.trusted_proxy?('127.0.0.6').must_equal true
+    req.trusted_proxy?('127.0.0.30').must_equal true
     req.trusted_proxy?('10.0.0.1').must_equal true
+    req.trusted_proxy?('10.000.000.001').must_equal true
     req.trusted_proxy?('172.16.0.1').must_equal true
     req.trusted_proxy?('172.20.0.1').must_equal true
     req.trusted_proxy?('172.30.0.1').must_equal true
     req.trusted_proxy?('172.31.0.1').must_equal true
+    req.trusted_proxy?('172.31.000.001').must_equal true
     req.trusted_proxy?('192.168.0.1').must_equal true
+    req.trusted_proxy?('192.168.000.001').must_equal true
     req.trusted_proxy?('::1').must_equal true
     req.trusted_proxy?('fd00::').must_equal true
+    req.trusted_proxy?('FD00::').must_equal true
     req.trusted_proxy?('localhost').must_equal true
     req.trusted_proxy?('unix').must_equal true
     req.trusted_proxy?('unix:/tmp/sock').must_equal true
@@ -1473,15 +1826,35 @@ EOF
     req.trusted_proxy?("unix.example.org").must_equal false
     req.trusted_proxy?("example.org\n127.0.0.1").must_equal false
     req.trusted_proxy?("127.0.0.1\nexample.org").must_equal false
+    req.trusted_proxy?("127.256.0.1").must_equal false
+    req.trusted_proxy?("127.0.256.1").must_equal false
+    req.trusted_proxy?("127.0.0.256").must_equal false
+    req.trusted_proxy?('127.0.0.300').must_equal false
+    req.trusted_proxy?("10.256.0.1").must_equal false
+    req.trusted_proxy?("10.0.256.1").must_equal false
+    req.trusted_proxy?("10.0.0.256").must_equal false
     req.trusted_proxy?("11.0.0.1").must_equal false
+    req.trusted_proxy?("11.000.000.001").must_equal false
     req.trusted_proxy?("172.15.0.1").must_equal false
     req.trusted_proxy?("172.32.0.1").must_equal false
+    req.trusted_proxy?("172.16.256.1").must_equal false
+    req.trusted_proxy?("172.16.0.256").must_equal false
     req.trusted_proxy?("2001:470:1f0b:18f8::1").must_equal false
   end
 
   it "sets the default session to an empty hash" do
     req = make_request(Rack::MockRequest.env_for("http://example.com:8080/"))
-    assert_equal Hash.new, req.session
+    session = req.session
+    assert_equal Hash.new, session
+    req.env['rack.session'].must_be_same_as session
+  end
+
+  it "sets the default session options to an empty hash" do
+    req = make_request(Rack::MockRequest.env_for("http://example.com:8080/"))
+    session_options = req.session_options
+    assert_equal Hash.new, session_options
+    req.env['rack.session.options'].must_be_same_as session_options
+    assert_equal Hash.new, req.session_options
   end
 
   class MyRequest < Rack::Request
@@ -1533,11 +1906,34 @@ EOF
     end
   }
 
+  (24..27).each do |exp|
+    length = 2**exp
+    it "handles ASCII NUL input of #{length} bytes" do
+      mr = Rack::MockRequest.env_for("/",
+        "REQUEST_METHOD" => 'POST',
+        :input => "\0"*length)
+      req = make_request mr
+      req.query_string.must_equal ""
+      req.GET.must_be :empty?
+      keys = req.POST.keys
+      keys.length.must_equal 1
+      keys.first.length.must_equal(length-1)
+      keys.first.must_equal("\0"*(length-1))
+    end
+  end
+
+  it "Env sets @env on initialization" do
+    c = Class.new do
+      include Rack::Request::Env
+    end
+    h = {}
+    c.new(h).env.must_be_same_as h
+  end
+
   class NonDelegate < Rack::Request
     def delegate?; false; end
-
     def query_parser
-      Rack::QueryParser.make_default(Rack::Utils.key_space_limit, Rack::Utils.param_depth_limit, bytesize_limit: 2**30)
+      Rack::QueryParser.make_default(Rack::Utils.param_depth_limit, bytesize_limit: 2**30)
     end
   end
 
@@ -1550,7 +1946,7 @@ EOF
       include Rack::Request::Helpers
       extend Forwardable
 
-      def_delegators :@req, :has_header?, :get_header, :fetch_header,
+      def_delegators :@req, :env, :has_header?, :get_header, :fetch_header,
         :each_header, :set_header, :add_header, :delete_header
 
       def_delegators :@req, :[], :[]=, :values_at
@@ -1561,10 +1957,8 @@ EOF
 
       def delegate?; true; end
 
-      def env; @req.env.dup; end
-
       def query_parser
-        Rack::QueryParser.make_default(Rack::Utils.key_space_limit, Rack::Utils.param_depth_limit, bytesize_limit: 2**30)
+        Rack::QueryParser.make_default(Rack::Utils.param_depth_limit, bytesize_limit: 2**30)
       end
     end
 
